# -*- coding: utf-8 -*-
#
# mimeparser/__init__.py
#
"""
MIME Parser package
"""
__docformat__ = "restructuredtext en"

__all__ = ('__version__', '__author__', '__email__', '__license__',
           '__credits__', 'MIMEParser',)

from .mimeparser import *


__version_info__ = {
    'major': 1,
    'minor': 0,
    'patch': 0,
    'releaselevel': 'final',
    'serial': 1
    }


def get_version(short=False):
    assert __version_info__['releaselevel'] in ('alpha', 'beta', 'final')
    vers = []
    vers.append("{major:d}".format(**__version_info__))
    vers.append("{minor:d}".format(**__version_info__))

    if __version_info__.get('patch', 0):
        vers.append("{patch:d}".format(**__version_info__))

    result = '.'.join(vers)

    if __version_info__.get('releaselevel') != 'final' and not short:
        result += "{}{:d}".format(
            __version_info__.get('releaselevel', 'a')[0],
            __version_info__.get('serial'))

    return result

__version__ = get_version()



# -*- coding: utf-8 -*-
#
# mimeparser/mimeparser.py
#
# See MIT License file.
#
"""
Parse MIME types that are usually found in HTTP headers.

Parsing `Accept` headers correctly has become very important with the
ubiquitus use of RESTful web services because, versioning of the service
is often defined in the mine type.

For reference see the following RFCs:

https://tools.ietf.org/html/rfc4288#section-3.2 (Vendor spec)
https://tools.ietf.org/html/rfc7231#section-5.3.1 (quality spec)
https://tools.ietf.org/html/rfc6839 (Suffix spec)

The basic idea of this code I got from Joe Gregorio,
https://github.com/jcgregorio/mimeparse

Entry point:
 - best_match() -- Primary method to find the mime type that would be
                   needed to find the closest match to the available mime
                   types.
 - parse_mime() -- Returns a parsed mime type into it's parts.
"""
__docformat__ = "restructuredtext en"
__author__ = 'Carl J. Nobile'
__email__ = 'carl.nobile@gmail.com'
__license__ = 'MIT License'
__credits__ = ''

__all__ = ('__docformat__', '__author__', '__email__', '__license__',
           '__credits__', 'MIMEParser',)

from decimal import *


class MIMEParser(object):

    def __init__(self):
        getcontext().prec = 3

    def best_match(self, available_mtypes, header_mtypes):
        """
        Return the best match from `header_mtypes` based on the
        `available_mtypes`.

        mtype is a value driectly from any header where mime types can be
        found. ie. Content-Type, Accept

        Examples:
          >>> best_match(['text/html, application/xbel+xml'],
                  'text/*;q=0.3, text/html;q=0.7, text/html;level=1,'
                  ' text/html;level=2;q=0.4, */*;q=0.5')
          
          >>> best_match(['application/xbel+xml', 'text/xml'],
                         'text/*;q=0.5,*/*; q=0.1')
          'text/xml'
        """
        weighted_matches = []
        pos = 0
        header_mtypes = [self.parse_mime(mt)
                         for mt in header_mtypes.split(',')]

        for mtype in available_mtypes:
            parsed_mtype = self.parse_mime(mtype)
            fit_and_q = self.__fitness_and_quality(parsed_mtype, header_mtypes)
            weighted_matches.append((fit_and_q, pos, mtype))
            pos += 1

        weighted_matches.sort()
        return weighted_matches[-1][0][1] and weighted_matches[-1][2] or ''

    def parse_mime(self, mtype):
        """
        Parses a mime-type into its component parts.

        Works with a single mime type and returns it's component parts in
        a tuple as in (type, subtype, suffix, params), where params is a
        dict of all the optional parameters of the mime type.

        For example, 'application/xhtml+xml;q=0.5;ver=1' would result in:

        ('application', 'xhtml', 'xml', {'q': Decimal(0.5),
                                         'ver': Decimal(1)}
        )

        All numeric values to any parameter are returned as a python
        Decimal object.
        """
        parts = mtype.split(';')
        params = {}

        # Split parameters and convert numeric values to a Decimal object.
        for k, v in [param.split('=', 1) for param in parts[1:]]:
            k = k.strip().lower()
            v = v.strip().lower()

            try:
                v = Decimal(v)
            except InvalidOperation:
                pass

            params[k] = v

        # Add/fix quality values.
        quality = params.get('q')

        if 'q' not in params or quality > Decimal(1) or quality < Decimal(0):
            params['q'] = Decimal(1)

        full_type = parts[0].strip().lower()

        # Fix non-standard single asterisk.
        if full_type == '*':
            full_type = '*/*'

        type, sep, subtype = full_type.partition('/')
        subtype, sep, suffix = subtype.rpartition('+')
        return type.strip(), subtype.strip(), suffix.strip(), params

    def __fitness_and_quality(self, preferred_mtype, ranges):
        """
        Find the best match for a pre-parsed preferred_mtype within
        pre-parsed ranges.

        Returns a tuple of the fitness value and the value of the 'q'
        (quality) parameter of the best match, or (-1, 0) if no match was
        found.
        """
        best_fit = -1
        best_fit_q = 0
        (target_type, target_subtype,
         target_suffix, target_params) = preferred_mtype

        for type, subtype, suffix, params in ranges:
            type_match = (type == target_type or type == '*'
                          or target_type == '*')
            subtype_match = (subtype == target_subtype or subtype == '*'
                             or target_subtype == '*')
            suffix_match = (suffix == target_suffix or suffix == ''
                            or not target_suffix)

            if type_match and subtype_match and suffix_match:
                param_matches = sum(
                    [1 for (key, value) in target_params.items()
                     if key != 'q' and key in params
                     and value == params[key]], 0)
                fitness = 1000 if type == target_type else 0
                fitness += 100 if subtype == target_subtype else 0
                fitness += 10 if suffix == target_suffix else 0
                fitness += param_matches

                if fitness > best_fit:
                    best_fit = fitness
                    best_fit_q = params.get('q', 0)

        return best_fit, Decimal(best_fit_q)



import os
import re
from setuptools import setup

def version():
    regex = r'^(?m){}[\s]*=[\s]*(?P<ver>\d*)$'

    with open(os.path.join(os.path.dirname(__file__), 'include.mk')) as f:
        ver = f.read()

    major = re.search(regex.format('MAJORVERSION'), ver).group('ver')
    minor = re.search(regex.format('MINORVERSION'), ver).group('ver')
    patch = re.search(regex.format('PATCHLEVEL'), ver).group('ver')
    return "{}.{}.{}".format(major, minor, patch)

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='parsing-tools',
    version=version(),
    packages=['mimeparser', 'xml2dict',],
    scripts=['bin/walker.py',],
    include_package_data=True,
    license='MIT License',
    description=('A set of utilities to help parse mime types found in '
                 'HTTP headers plus a utility to parse XML documents into '
                 'Python objects.'),
    long_description=README,
    url='https://github.com/cnobile2012/python_tools/',
    author='Carl J. Nobile',
    author_email='carl.nobile@gmail.com',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        ],
    install_requires=[
        'defusedxml',
        'six',
        ],
    )



# -*- coding: utf-8 -*-
#
# xml2dict/xml2dict.py
#
# See MIT License file.
#
"""
Convert XML to a Python dict.

"""
__docformat__ = "restructuredtext en"
__author__ = 'Carl J. Nobile'
__email__ = 'carl.nobile@gmail.com'
__license__ = 'MIT License'
__credits__ = ''

__all__ = ('__docformat__', '__author__', '__email__', '__license__',
           '__credits__', 'XML2Dict',)

import io
import re
import six
import logging
import defusedxml.ElementTree as ET


class XML2Dict(object):
    __NSPACE_REGEX = r"^\{(?P<uri>.*)\}(?P<tag>.*)$"
    __NSPACE_OBJ = re.compile(__NSPACE_REGEX)

    def __init__(self, empty_tags=True, rm_whitespace=True, logger_name='',
                 level=None):
        if logger_name == '':
            logging.basicConfig()

        self._log = logging.getLogger(logger_name)

        if level:
            self._log.setLevel(level)

        self.__namespaces = {}
        self.__empty_tags = empty_tags
        self.__rm_whitespace = rm_whitespace

    def _set_file_object(self, xml):
        if isinstance(xml, io.IOBase):
            xml.seek(0) # Make sure we're at the start of the file.
            self._xml = xml
        else:
            self._xml = six.StringIO(xml)

    def parse(self, xml, encoding=None):
        data = []
        parser = ET.DefusedXMLParser(encoding=encoding)
        self._set_file_object(xml)

        try:
            tree = ET.parse(self._xml, parser=parser, forbid_dtd=True)
            root = tree.getroot()
            self.__namespaces.update(dict([(v,k) for k,v in root.items()]))
            self.__node(data, root)
        except ET.ParseError as e:
            self._log.error("Could not parse xml, %s", e)
            raise e

        self._log.debug("data: %s", data)
        return data

    def __node(self, data, node):
        child_data = {}
        data.append(child_data)
        # Process tag
        tag_name = node.tag
        nspace, name = self.__split_namespace(tag_name)
        ns_tag = "{}{}{}".format(nspace, ':' if nspace else '', name)
        text = self.value_hook(node.text)
        child_data['attrib'] = {k: self.value_hook(v)
                                for k,v in node.attrib.items()}
        child_data['tag'] = {'nspace': nspace,
                             'tag': name,
                             'value': self.__tag_value(text)}
        children_data = []
        child_data['children'] = children_data

        for child in node.getchildren():
            self.__node(children_data, child)

    def __split_namespace(self, tag):
        sre = self.__NSPACE_OBJ.search(tag)

        if sre:
            nspace = sre.group('uri')
            nspace = self.__namespaces.get(nspace, '')
            name = sre.group('tag')
        else:
            nspace = ''
            name = tag

        return nspace, name

    def __tag_value(self, text):
        if text:
            if self.__rm_whitespace:
                text = text.strip()
        elif self.__empty_tags:
            text = ''

        return text

    def value_hook(self, value):
        """
        This hook can be overridden to convert values to Python types.
        """
        return value

def _openFile(filePath, how='rb'):
    fp = None

    try:
        fp = open(filePath, how)
    except IOError as e:
        msg = "Error: Could not open file: {}, {}."
        raise IOError(msg.format(filePath, e))

    return fp


if __name__ == '__main__':
    import sys, traceback
    from pprint import pprint

    # encoding="utf-8"
    data = '''<?xml version="1.0" standalone="yes" ?>
<response status="ok">
  <user name="example.gen"
        type="Generic"
        descr="Example User"
        homedirectory=""
        homedrive=""
        loginscript=""
        profilepath=""
        loginshell=""
        unixhomedirectory=""
        env="dev.example.com"
        status="enabled"
        passwordexpires="2015-10-21T21:16:14Z"
        passwordlastset="2015-04-24T21:16:14Z"
        dn="CN=test-root.gen,OU=Generics,OU=Example Users,DC=dev,DC=example,DC=com"
        upn="test-root.gen@dev.example.com"
        url="https://ds-api.example.com/DS/v1/dev.example.com/User/test-root.gen">
    <env-list count="2">
      <env>dev.example.com</env>
      <env>dsxdev.example.com</env>
    </env-list>
    <owner-list count="2">
      <owner>joeblow</owner>
      <owner>jhamilton</owner>
    </owner-list>
  </user>
</response>
'''

    try:
        if len(sys.argv) > 1:
            xml = _openFile(sys.argv[1])
        else:
            xml = data

        x2d = XML2Dict(level=logging.DEBUG)
        out = x2d.parse(xml)
        print("\nResults:")
        pprint(out)
    except:
        tb = sys.exc_info()[2]
        traceback.print_tb(tb)
        print("{}: {}\n".format(sys.exc_info()[0], sys.exc_info()[1]))
        sys.exit(1)

    sys.exit(0)
